/* This program is Copyright (c) 1990 David Allen.  It may be freely
   distributed as long as you leave my name and copyright notice on it.
   I'd really like your comments and feedback; send e-mail to
   davea@vlsi.ll.mit.edu, or send us-mail to David Allen, 10 O'Moore Ave,
   Maynard, MA 01754. */

/* This file contains the function getparams(), which reads parameters
   from an input file.  The default values for all of the parameters are
   set here as well. */

#include "const.h"

/* These are all of the adjustable parameters; they are completely
   described in the file params.doc. */

long XSIZE = 90,	YSIZE = 90,	MAXSTEP = 100;
long MAXBUMP = 50,	BUMPTOL = 50;
long DRAWEVERY = 1,	BLOBLEVEL = 64,	DRAWMODE = DRAWMODE_GENERIC;
long ZINIT = 22,	ZSUBSUME = 16,	ZCOAST = 16;
long ZSHELF = 8,	ZMOUNTAIN = 48;
long RIFTPCT = 40,	DOERODE = 1,	ERODERND = 4;
long MAXCTRTRY = 50,	RIFTDIST = 5,	BENDEVERY = 6;
long BENDBY = 50,	SPEEDRNG = 300,	SPEEDBASE = 200;

double MR [] = { 1.0,1.0,1.0,0.7,0.4,0.1,-0.2,-0.5,-0.8,-1.0 };
long MAXLIFE = 10; /* Length of MR vector */


#include <stdio.h>
FILE *fp;

/* The input is handled by one function, gettoken, which returns one
   of the following codes each time it is called. */
#define TOK_EOF 0
#define TOK_OPEN 1
#define TOK_CLOSE 2
#define TOK_WORD 3

short linecount = 1; /* So error messages can refer to a line number */
char getbuf[256]; /* Storage for each token as it is being read */

/* A shorthand for string comparison */
#define CMP(x) (!strcmp (getbuf, x))

extern float atof ();


getparams (s) char *s; { 
   /* This function is called by init() with one parameter, a string which
   comes directly off the command line.  If the string is null, getparams()
   is never called; otherwise, a "-" means standard input, while any other
   text is treated as a filename and handed to fopen().  If the file doesn't
   exist the program exits via panic().  Once the file is open, the function
   expects to find paren-delimited pairs of (name value).  Value could be
   a vector, like (name (val1 val2)).  After a name is found, there is
   a very simple test, one string comparison per adjustable parameter,
   to see if the name is recognized.  If unrecognized, the program panics.
   If it is a recognized name, the appropriate parsing function, with the
   appropriate parameters, is called.  There are three parsing functions,
   getdim(), getlng(), and getdvec(). */

   short x, i, j;

   if (!strcmp (s, "-")) fp = stdin;
   else { fp = fopen (s, "r"); if (!fp) panic ("Can't find input file"); }
   
   while ((x = gettoken (getbuf)) != TOK_EOF) {
      if (x != TOK_OPEN) geterr ("expected open paren");
      if (gettoken (getbuf) != TOK_WORD) geterr ("expected parameter name");
      if      CMP ("XSIZE")	getdim  (&XSIZE);
      else if CMP ("YSIZE")	getdim  (&YSIZE);
      else if CMP ("MOVERATE")	getdvec (&MAXLIFE, MR);
      else if CMP ("MAXSTEP")	getlng  (&MAXSTEP);
      else if CMP ("MAXBUMP")	getlng  (&MAXBUMP);
      else if CMP ("BUMPTOL")	getlng  (&BUMPTOL);
      else if CMP ("DRAWEVERY")	getlng  (&DRAWEVERY);
      else if CMP ("DRAWMODE")	getlng  (&DRAWMODE);
      else if CMP ("BLOBLEVEL")	getlng  (&BLOBLEVEL);
      else if CMP ("ZINIT")	getlng  (&ZINIT);
      else if CMP ("ZSUBSUME")	getlng  (&ZSUBSUME);
      else if CMP ("ZCOAST")	getlng  (&ZCOAST);
      else if CMP ("ZSHELF")	getlng  (&ZSHELF);
      else if CMP ("ZMOUNTAIN")	getlng  (&ZMOUNTAIN);
      else if CMP ("RIFTPCT")	getlng  (&RIFTPCT);
      else if CMP ("DOERODE")	getlng  (&DOERODE);
      else if CMP ("ERODERND")	getlng  (&ERODERND);
      else if CMP ("MAXCTRTRY")	getlng  (&MAXCTRTRY);
      else if CMP ("RIFTDIST")	getlng  (&RIFTDIST);
      else if CMP ("BENDEVERY")	getlng  (&BENDEVERY);
      else if CMP ("BENDBY")	getlng  (&BENDBY);
      else if CMP ("SPEEDBASE")	getlng  (&SPEEDBASE);
      else if CMP ("SPEEDRNG")	getlng  (&SPEEDRNG);
      else geterr ("unknown parameter name");
      if (gettoken (getbuf) != TOK_CLOSE) geterr ("expected close paren"); } }


gettoken (s) char *s; {
   /* This is the only function which actually reads from the file.  It
   maintains a one-character unget buffer.  It ignores initial whitespace,
   but counts newlines to maintain an accurate linecount.  Open or close
   parentheses count as tokens, and so does any amount of text with no
   whitespace or parens.  The return code indicates whether end of file,
   open paren, close paren, or a word were found.  If a word was found, it
   is copied into the string pointed to by s.  When a word is found, the
   character which terminated the word (whitespace, EOF or paren) is put
   into the unget buffer to be read the next time gettoken() is called. */

   static char buf = 0; char c;

   white: if (buf) { c = buf; buf = 0; } else c = getc (fp);
   switch (c) {
      case '\n': linecount++;
      case '\t': 
      case ' ' : goto white; break;
      case EOF : return (TOK_EOF); break;
      case '(' : return (TOK_OPEN); break;
      case ')' : return (TOK_CLOSE); break; }
   text: if ((c==' ')||(c=='\t')||(c=='\n')||(c=='(')||(c==')')||(c==EOF)) {
      buf = c; *s = 0; return (TOK_WORD); }
   else { *s++ = c; c = getc (fp); goto text; } }


geterr (s) char *s; {
   /* Calls machine-specific panic() with nicely formatted message */
   sprintf (getbuf, "Parameter error: %s in line %d", s, linecount);
   panic (getbuf); }


getlng (x) long *x; {
   /* Called after reading a name associated with a single long,
   this function just reads a long from the input file and uses atoi()
   to convert it into a long, stored at the address passed in.   Note
   that any text, including a string or a real, will be read; no type
   checking is performed. */

   if (gettoken (getbuf) != TOK_WORD) geterr ("expected long");
   *x = atoi (getbuf); }


getdim (x) long *x; {
   /* Called right after reading a name associated with an array bound,
   this function reads one long from the input file; if the value is
   greater than the default value assigned above, the user is trying
   to exceed a compiled-in limit.  That's an error. */

   long y;

   if (gettoken (getbuf) != TOK_WORD) geterr ("expected long");
   y = atoi (getbuf);
   if (y > *x) geterr ("dimension exceeds reserved space");
   *x = y; }


getdvec (dim, v) long *dim; double *v; {
   /* Called right after reading a name associated with a one-dimensional
   array of doubles, this function expects to find a list of doubles
   starting with an open paren and ended by a closing paren.  The parameter
   dim contains the compiled-in array limit; if the input file contains
   more than this number of entries, the function complains.  No type checking
   is done; atof() is used to convert any text to a double.  On exit, the
   dimension is set to the correct value. */

   short x; long i = 0;

   if (gettoken (getbuf) != TOK_OPEN)
      geterr ("expected open paren");
   while ((x = gettoken (getbuf)) == TOK_WORD) {
      if (i == *dim) geterr ("vector is too long");
      v[i++] = atof (getbuf); }
   if (x != TOK_CLOSE) geterr ("expected close paren");
   *dim = i; } 
